
local addon_name, addon = ...

local module = addon:NewModule("Ready", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
addon.Ready = module

local LibPetJournal = LibStub("LibPetJournal-2.0")

--
--
--

local IsFalling, IsFlying, IsStealthed = IsFalling, IsFlying, IsStealthed
local GetTime, IsMounted, UnitAura = GetTime, IsMounted, UnitAura
local UnitCastingInfo, UnitChannelInfo = UnitCastingInfo, UnitChannelInfo
local tconcat, tinsert = table.concat, table.insert
local GetSummonedPetID = C_PetJournal.GetSummonedPetGUID

--
--
--

local INVIS_SPELLS = {66, 11392, 3680, 80326}

--
--
--

local SPELL_NAME_CACHE = setmetatable({}, {
    __index = function(self, key)
        local name = GetSpellInfo(key)
        rawset(self, key, name)
        return name
    end
})

function module:OnInitialize()
    self.is_true = {}                           -- check that is not preventing summon
    self.is_false = { addon_ready = true }      -- check that is preventing summon
    self.is_irrelevant = {}                     -- disabled checks
    self.desummon_on = {}                       -- if these checks are false, desummon

    self.last_spell_sent = 0
    
    self:RegisterEvent("COMPANION_UPDATE")
    self:RegisterEvent("UPDATE_STEALTH")
    self:RegisterEvent("UNIT_AURA")
    self:RegisterEvent("UNIT_SPELLCAST_SENT")
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "SpellCastUpdate")
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "SpellCastUpdate")
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE", "SpellCastUpdate")
    self:RegisterEvent("UNIT_SPELLCAST_FAILED", "SpellCastUpdate")
    self:RegisterEvent("UNIT_SPELLCAST_START", "SpellCastUpdate")
    self:RegisterEvent("UNIT_SPELLCAST_STOP", "SpellCastUpdate")
    self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "SpellCastUpdate")
    self:RegisterEvent("PLAYER_REGEN_DISABLED")
    self:RegisterEvent("PLAYER_REGEN_ENABLED")
    self:RegisterEvent("PLAYER_UPDATE_RESTING")
    self:RegisterEvent("BARBER_SHOP_CLOSE")
    self:RegisterEvent("BARBER_SHOP_OPEN")
    self:RegisterEvent("LOOT_OPENED")
    self:RegisterEvent("LOOT_CLOSED")
    self:RegisterEvent("PLAYER_ALIVE")
    self:RegisterEvent("PLAYER_DEAD")
    self:RegisterEvent("PLAYER_UNGHOST")
    self:RegisterEvent("CONFIRM_XP_LOSS")
    self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
    self:RegisterEvent("UNIT_ENTERED_VEHICLE")
    self:RegisterEvent("UNIT_EXITED_VEHICLE")
    
    self:UpdateChecks()
end

function module:OnEnable()
    self:PetListUpdated()
    self:ScheduleRepeatingTimer("FrequentCheck", 0.1)
    self:SetReady("addon_enabled", addon:IsEnabledSummoning())
    self:SetReady("combat", not InCombatLockdown())
    self:SetReady("resting", IsResting())
    self:SetReady("alive", not UnitIsDeadOrGhost("player"))
    self:SetReady("in_battleground", not addon.InBattlegroundOrArena())
    self:SetReady("vehicle", not UnitInVehicle("player"))
    self:UNIT_AURA(nil, "player")
    self:UPDATE_STEALTH()

    self:SetReady("addon_ready", true)

    self:StartNearPEW()
    self:RegisterEvent("PLAYER_ENTERING_WORLD", "StartNearPEW")
    
    LibPetJournal.RegisterCallback(self, "PetListUpdated")
    self:RegisterMessage("PetLeash-EnableState", "OnAddOnState")
end

function module:OnAddOnState(event, enabled)
    self:SetReady("addon_enabled", enabled)
end

local function debug_ready_name(module, name, irr)
    local flag = {}
    if module.desummon_on[name] then
        tinsert(flag, "UN")
    end
    if irr then
        tinsert(flag, "IGN")
    end
    if #flag > 0 then
        return ("%s[%s]"):format(name, tconcat(flag, ","))
    else
        return name
    end
end

function module:DumpDebugState()
    local ready = {}
    local not_ready = {}
    
    for name,_ in pairs(self.is_true) do
        tinsert(ready, debug_ready_name(self, name))
    end
    for name,_ in pairs(self.is_false) do
        tinsert(not_ready, debug_ready_name(self, name))
    end
    for name,value in pairs(self.is_irrelevant) do
        if value then
            tinsert(ready, debug_ready_name(self, name, true))
        else
            tinsert(not_ready, debug_ready_name(self, name, true))
        end
    end

    table.sort(ready)
    table.sort(not_ready)
    
    self:Printf("Ready: %s", tconcat(ready, " "))
    self:Printf("Not Ready: %s", tconcat(not_ready, " "))
    self:Printf("Last: %d %d, sum=%s, cast=%f",
                IsFalling(), HasFullControl(),
                GetSummonedPetID() or "nil",
                GetTime() - self.last_spell_sent)
end

function module:UpdateChecks(profile)
    profile = profile or addon.db.profile

    self:EnableCheck("addon_ready", true)
    self:EnableCheck("combat", true)
    self:EnableCheck("in_battleground",
            profile.dismissInBattleground or not profile.enableInBattleground,
            profile.dismissInBattleground)
    self:EnableCheck("resting", profile.disableOutsideCities)
    self:EnableCheck("stealthed", true, profile.dismissWhenStealthed)
    self:EnableCheck("invisible", true, profile.dismissWhenStealthed)
    self:EnableCheck("flying", true, profile.dismissWhileFlying)
    self:EnableCheck("near_pew", profile.nearPEWOnly)
    self:EnableCheck("alive", true)
    self:EnableCheck("casting", true)
    self:EnableCheck("have_pets", true)
    self:EnableCheck("barbar", true)
    self:EnableCheck("thunderisle_saurok", true)
    self:EnableCheck("eating", true)
    self:EnableCheck("feigned", true)
    self:EnableCheck("looting", true)
    self:EnableCheck("mounted", true)
    self:EnableCheck("vehicle", true)           -- XXX not tested
    self:EnableCheck("eating", true)
    self:EnableCheck("addon_enabled", true)
    
    self:_CheckSummon()
end

function module:EnableCheck(name, enabled, desummon)
    local v = true
    if self.is_irrelevant[name] ~= nil then
        v = self.is_irrelevant[name]
        self.is_irrelevant[name] = nil
    elseif self.is_true[name] then
        v = true
        self.is_true[name] = nil
    elseif self.is_false[name] then
        v = false
        self.is_false[name] = nil
    end

    if enabled then
        if v then
            self.is_true[name] = true
        else
            self.is_false[name] = true
        end
        
        if desummon then
            self.desummon_on[name] = true
        else
            self.desummon_on[name] = nil
        end
    else
        self.is_irrelevant[name] = v
    end
end

function module:SetReady(name, value)
    if value then
        if self.is_false[name] then
             self.is_false[name] = nil
             self.is_true[name] = true
             self:_CheckSummon()
             return
        elseif self.is_true[name] then
            return
        end
    else
        if self.is_true[name] then
            self.is_true[name] = nil
            self.is_false[name] = true
            self:_CheckUn(name)
            return
        elseif self.is_false[name] then
            return
        end
    end
    
    self.is_irrelevant[name] = not not value
end

function module:LastCheck()
    -- these checks are not tied to an event
    return not IsFalling()
        and HasFullControl()
        and GetTime() - self.last_spell_sent > 1.5
end

function module:_CheckSummon(wait_timer_finished)
    if not next(self.is_false) then
        if addon:HasPet() and not self.force then
            -- nothing to do
            return
        end
        
        if wait_timer_finished then
            if self:LastCheck() then
                if self.force then
                    -- XXX We used to desummon our pet here, but currently
                    -- that just causes SummonPetByID to ignore our choice
                    -- and summon something completely at random.
                    self.force = false
                end
                addon:SummonPet()
                self:ScheduleTimer("CheckForSuccess", 2)
                return
            end
        elseif self.wait_timer then
            return
        end
        
        self.wait_timer_counter = 0
        self.wait_timer = self:ScheduleRepeatingTimer("WaitTimerProgress", 0.5)
    end
end

function module:_CheckUn(name)
    --- TODO add a special flag to unsummon even when we are not enabled
    if self.desummon_on[name] and addon:IsEnabledSummoning() then
        if not InCombatLockdown() then
            addon:DesummonPet(false, true)
        end
    end
    
    if self.wait_timer then
        self:CancelTimer(self.wait_timer)
        self.wait_timer = nil
    end
end

function module:WaitTimerProgress()
    self.wait_timer_counter = self.wait_timer_counter + 0.5
    if self.wait_timer_counter >= addon.db.profile.waitTimer then
        self:CancelTimer(self.wait_timer)
        self.wait_timer = nil
        self:_CheckSummon(true)
    elseif not self:LastCheck() then
        self.wait_timer_counter = 0             -- reset
    end
end

function module:CheckForSuccess()
    self:_CheckSummon()
end

function module:QueueResummon(nowait)
    -- if nowait is true, we pretend we are coming from the timer
    self.force = true
    self:_CheckSummon(nowait)
end

function module:PetListUpdated()
    self:SetReady("have_pets", LibPetJournal:NumPets())
end

function module:FrequentCheck()
    self:SetReady("flying", not IsFlying())
end

function module:StartNearPEW()
    if not self.near_pew_handle then
        self:SetReady("near_pew", true)
        self.near_pew_handle = self:ScheduleTimer("NearPEWFinish", 6 + (addon.db.profile.waitTimer or 1))
    end
end

function module:NearPEWFinish()
    self:SetReady("near_pew", false)
    self.near_pew_handle = false
end

function module:COMPANION_UPDATE(event, ctype)
    if ctype == "CRITTER" then
        self:_CheckSummon()
    end
end
 
function module:BARBER_SHOP_OPEN()
    self:SetReady("barbar", false)
end

function module:BARBER_SHOP_CLOSE()
    self:SetReady("barbar", true)
end

function module:LOOT_OPENED()
    self:SetReady("looting", false)
end

function module:LOOT_CLOSED()
    self:SetReady("looting", true)
end

function module:PLAYER_ALIVE()
    self:SetReady("alive", not UnitIsDeadOrGhost("player"))
end

function module:PLAYER_DEAD()
    self:SetReady("alive", false)
end

function module:PLAYER_UNGHOST()
    self:SetReady("alive", true)
end

function module:CONFIRM_XP_LOSS()
    self:SetReady("alive", true)
end

function module:UPDATE_STEALTH()
    if IsStealthed() then
        self:SetReady("stealthed", false)
    else
        self:SetReady("stealthed", true)
    end
end

function module:UNIT_AURA(event, unit)
    if unit ~= "player" then
        return
    end
   
    -- check for invisibility
    local invisible
    for i,spellid in ipairs(INVIS_SPELLS) do
        if UnitAura("player", SPELL_NAME_CACHE[spellid], nil, "HELPFUL") then
            invisible = true
            break
        end
    end
   
    if(invisible) then
        self:SetReady("invisible", false)
    else
        self:SetReady("invisible", true)
    end

    -- sauroked
    self:SetReady("thunderisle_saurok",
                not UnitAura("player", SPELL_NAME_CACHE[136461], nil, "HELPFUL"))
   
    -- eating means we are sitting down
    if UnitAura("player", SPELL_NAME_CACHE[430], nil, "HELPFUL") then
        self:SetReady("eating", false)
    elseif UnitAura("player", SPELL_NAME_CACHE[433], nil, "HELPFUL") then
        self:SetReady("eating", false)
    else
        self:SetReady("eating", true)
    end

    -- feign death
    if UnitAura("player", SPELL_NAME_CACHE[5384], nil, "HELPFUL") then
        self:SetReady("feigned", false)
    else
        self:SetReady("feigned", true)
    end
    
    self:ScheduleTimer("PostAuraCheck", 0)
end

function module:PostAuraCheck()
    self:SetReady("mounted", not IsMounted())
end

function module:UNIT_SPELLCAST_SENT(event, unit)
    if unit == "player" then
        self.last_spell_sent = GetTime()
        return self:SpellCastUpdate(event, unit)
    end
end

function module:SpellCastUpdate(event, unit)
    if unit == "player" then
        self:SetReady("casting", not (UnitCastingInfo("player") or UnitChannelInfo("player")))
    end
end

function module:PLAYER_REGEN_DISABLED()
    self:SetReady("combat", false)
end

function module:PLAYER_REGEN_ENABLED()
    self:SetReady("combat", true)
end

function module:PLAYER_UPDATE_RESTING()
   self:SetReady("resting", IsResting())
end

function module:ZONE_CHANGED_NEW_AREA()
    self:SetReady("in_battleground", not addon.InBattlegroundOrArena())
end

function module:UNIT_ENTERED_VEHICLE(event, unit)
    if unit == "player" then
        self:SetReady("vehicle", false)
    end
end

function module:UNIT_EXITED_VEHICLE(event, unit)
    if unit == "player" then
        self:SetReady("vehicle", true)
    end
end

